$CACHE = 1;

sub cache_init{
    local(*FH);
    my $cache = "$DbPath.cache.0";
    if (!$CacheInit){
        $CacheInit = 1;
        if (-s $cache && (stat(_))[9] < $HashTime){
            if (-w $cache){
                local(*TMPFH);
                if (open(FH, $cache)){
                    my $str;
                    while (defined($str = <FH>)){
                        my ($ext, $elem) = split(/ /, $str, 2);
                        open(TMPFH, ">$DbPath.cache.$ext");
                        close TMPFH;
                    }
                    close(FH);
                }
                open(TMPFH, ">$cache");
                close TMPFH;
            }
            return if -s $cache;
        }
        if (open(FH, $cache)){
            my $str;
            while (defined($str = <FH>)){
                chomp $str;
                my ($ext, $elem) = split(/ /, $str, 2);
                my $cfile = "$DbPath.cache.$ext";

                if (-r $cfile && -s $cfile){
                    $CacheSize{$ext} = (stat(_))[7];
                    $CacheTim{$ext} = (stat(_))[9];

                    $CacheList{$elem} = $ext;
                    $CacheInvList{$ext} = $elem;
                }
            }
            close(FH);
        }
    }
}

sub cache_write_list{
    local(*FH);
    my $umask = umask 0;

    my $size = 0;
    for $ext (keys %CacheInvList){
        $size += $CacheSize{$ext};
    }

    if ($size > $CacheSizeLimit){
        &debug_log("Cache size over");
        for $ext (sort {($CacheTim{$a} <=> $CacheTim{$b})
                            || ($CacheSize{$b} <=> $CacheSize{$a})
                            } keys %CacheInvList){
            my $file = "$DbPath.cache.$ext";
            next unless -w $file;
            local(*TMPFH);
            next unless open(TMPFH, ">$file");
            close TMPFH;
            $size -= $CacheSize{$ext};
            delete $CacheInvList{$ext};
            &debug_log("Cache remove: $file");
            last if $size <= $CacheSizeLimit;
        }
    }

    if (open(FH, ">$DbPath.cache.0")){
        for $ext (keys %CacheInvList){
            print FH "$ext $CacheInvList{$ext}\n";
        }
        close(FH);
        &debug_log("Cache write list: $ext");
    }
    umask $umask;
}

sub cache_close{
    undef %CacheInvList;
    undef %CacheSize;
    undef %CacheTim;
    undef %CacheList;
    undef %CacheInvList;
    $CacheInit = 0;
    $CacheChange = 0;
}

sub cache_read_proc{
    local($field, *score, $fptr, @para) = @_;
    my $ext = $CacheList{$field};
    if ($ext){
        my $file = "$DbPath.cache.$ext";
        local(*FH);
        if (&openbfile(FH, $file)){
            my($buf, $hit, $ndx);
            read(FH, $buf, $CacheSize{$ext});
            close(FH);
            unless (utime $^T, $^T, $file){
                &debug_log("Cache utime failed: $file");
                if (open(FH, ">$file")){
                    print FH $buf;
                    close(FH);
                    &debug_log("Cache rewrite: $file");
                }
            }
            $CacheTim{$ext} = $^T;

            for $ndx (&unpackw($buf)){
                &$fptr(\$hit, $ndx, *score, @para);
            }
            &debug_log("Cache read: $file");
            return $hit;
        }else{
            delete $CacheList{$field};
            delete $CacheInvList{$ext};
            $CacheChange = 1;
        }
    }
    return -1;
}

sub cache_read_word{
    local($field, *score, @para) = @_;
    return &cache_read_proc($field, *score, \&ssub, @para);
}

sub cache_read_sub{
    local($phit, $ndx, *score) = @_;
    $score{$ndx} = -1, ++$$phit if &TimEnable($ndx);
}

sub cache_read_score{
    local($field, *score) = @_;
    return &cache_read_proc($field, *score, \&cache_read_sub);
}

sub cache_write_proc{
    my ($field, @ndx) = @_;
    chomp $field;
    local(*FH);
    my $umask = umask 0;
    my $ext = $CacheList{$field};
    unless ($ext){
        $ext = 1;
        $ext++ while $CacheInvList{$ext};
        $CacheList{$field} = $ext;
        $CacheInvList{$ext} = $field
    }

    my $file = "$DbPath.cache.$ext";
    if (open(FH, ">$file")){
        binmode(FH);
        $CacheSize{$ext} = 0;
        for $ndx (@ndx){
            print FH pack($IntType, $ndx);
            $CacheSize{$ext} += $IntSize;
        }
        close(FH);

        utime $^T, $^T, $file;
        $CacheTim{$ext} = $^T;

        $CacheChange = 1;

        &debug_log("Cache write: $file");
    }else{
        delete $CacheList{$field};
        delete $CacheInvList{$ext};
    }
    umask $umask;
}

sub cache_write_word{
    &cache_write_proc(@_);
}

sub cache_write_score{
    local($field, *score) = @_;
    &cache_write_proc($field, sort {$a <=> $b} keys %score);
    return;
}

