#! /usr/local/bin/perl5 -w
# -*- Perl -*-
#
# nmzseparate.pl - program to separate index
#         by furukawa@tcp-ip.or.jp
#

use strict;

push(@INC, "/usr/local/share/namazu/pl");
{
    my($tmp) = $0;
    $tmp = "./$tmp" if $tmp !~ /[\/\\]/;
    while ($tmp =~ /^(.*[\/\\])/){
        unshift(@INC, $1);
        last if !-l $tmp;
        $tmp = $1 . $tmp if ($tmp = readlink($tmp)) !~ /^\//;
    }
}
require 'nmzidx.pl';

&usage, exit if @ARGV != 4;

&nmzseparate(@ARGV);

sub nmzseparate{
    my ($pattern, $dir0, $dir1, $dir2) = @_;

    my $src = new nmzidx($dir0, 'r');

    my @dst = (new nmzidx($dir2, 'w'), new nmzidx($dir1, 'w'));

    my @separate = ();
    {
        my $src_file = $src->open_flist;
        my @field = keys %{$src_file->{'field'}};

        my @dst_file = map {$_->open_flist} @dst;

        my $evalstr = '';
        $evalstr .= "my \$r = \$list{'r'};\n";
        $evalstr .= "my \$t = \$list{'t'};\n";

        for my $field (@field){
            my $tmp = $field;
            $tmp =~ s/[^A-Za-z]+//g;
            $evalstr .= "my \$$tmp = \$list{'field'}{'$field'};\n";
            for my $dst (@dst){
                $dst->{'flist'}->{'field'}->open($dst, $field) unless defined $dst->{'flist'}->{'field'}->{$field};
            }
        }
        
        $evalstr .= "\$ret = ($pattern);\n";

        my @size = (0, 0);
        my %list;
        my $ndx = 0;
        my $size = $src_file->{'size'};
        while (defined $src_file->read(\%list)){
            my $ret;
            eval($evalstr);
            my $separate = $ret? 1: 0;
            push(@separate, {'separate' => $separate,
                             'no' => $size[$separate]});
            ++$size[$separate];
            $dst[$separate]->{'flist'}->write(\%list);
            print "$ndx/$size\n" unless ++$ndx % 100;
        }
        $src_file->close;
        map {$_->close} @dst_file;
    }
    {
        my $src_word = $src->open_word;

        my @dst_word = map {$_->open_word} @dst;

        my $ndx = 0;
        my $size = $src_word->{'size'};

        my ($word, %list);
        while (defined $src_word->read(\$word, \%list)){
            my(%list1, %list2);
            my @refs = (\%list2, \%list1);

            for my $key (keys %list){
                my $separate = $separate[$key]->{'separate'};
                my $no = $separate[$key]->{'no'};
                $refs[$separate]->{$no} = $list{$key};
            }

            for my $key (0, 1){
                if (keys %{$refs[$key]}){
                    $dst_word[$key]->write($word, $refs[$key]);
                }
            }
            print "$ndx/$size $word\n" unless ++$ndx % 100;
        }
        $src_word->close;
        map {$_->close} @dst_word;
    }
    {
        my $src_phrase = $src->open_phrase;
        my @dst_phrase = map {$_->open_phrase} @dst;

        for (my $ndx = 0; $ndx < 0x10000; $ndx++){
            my (@list, @list1, @list2);
            my @refs = (\@list2, \@list1);

            $src_phrase->read(\@list);
            for my $key (@list){
                my $separate = $separate[$key]->{'separate'};
                my $no = $separate[$key]->{'no'};
                push(@{$refs[$separate]}, $no);
            }
            for my $ndx (0, 1){
                $dst_phrase[$ndx]->write($refs[$ndx]);
            }
            printf("phrase %04X\n", $ndx) unless $ndx & 0xff;
        }
    }

    for my $dst (@dst){
        $dst->write_status($src);
        if (my $log = $dst->log_open("[Separate]")){
            $log->printf("%-20s %d\n", "Total Files:", $dst->{'flist'}->{'offset'});
            $log->printf("%-20s %d\n", "Total Keywords:", $dst->{'word'}->{'offset'});
            $dst->log_close;
        }
        $dst->replace_db(0);
    }
}

sub usage{
    print("Usage: nmzseparate.pl pattern src dst1 dst2\n");
}
